import dataclasses
import json
import logging
from datetime import datetime
from pathlib import Path
from typing import Any, Iterable, Iterator, Sequence, cast

from pandas import DataFrame

from aau_label.errors import ClassFileNotFoundError

from . import utilities
from .io import Darknet, Pascal
from .model import AAULabel, AAULabelImage, COCOInfo, COCOLicense
from .protocols import Label, LabelImage, LabelImageDeserializer

logger = logging.getLogger(__name__)


class DeserializedLabelImageIterator(Iterator[LabelImage]):
    def __init__(
        self,
        img_label_pairs: Iterator[tuple[Path, Path]],
        deserializer: LabelImageDeserializer,
    ) -> None:
        self.__pairs = img_label_pairs
        self.__deserializer = deserializer

    def __iter__(self):
        return self

    def __next__(self) -> LabelImage:
        img_path, label_path = next(self.__pairs)
        return self.__deserializer.deserialize(img_path, label_path)


def from_file(filepath: str | Path) -> list[LabelImage]:
    """
    Loads a list of LabelImage objects from a AAU JSON file.

    Args:
        filepath (str | Path): The path to the AAU JSON file to load from.

    Returns:
        list[LabelImage]: A list of LabelImage objects loaded from the file.
    """
    label_images = []
    with open(filepath, "r") as file:
        data: list[dict[str, Any]] = json.load(file)

        for element in data:
            labels: list[Label] = []
            for label_dict in element["labels"]:
                label_dict = cast(dict[str, Any], label_dict)
                # Old format uses classifier, new format uses name
                label_name = label_dict.get("classifier", label_dict["name"])

                label = AAULabel(
                    label_dict["x"],
                    label_dict["y"],
                    label_dict["width"],
                    label_dict["height"],
                    label_name,
                )
                labels.append(label)

            label_image = AAULabelImage(
                path=Path(element["path"]),
                width=element["width"],
                height=element["height"],
                labels=labels,
            )
            label_images.append(label_image)

    return label_images


def __get_files_with_extensions(
    directory: Path, extensions: set[str]
) -> Iterator[Path]:
    for file in directory.iterdir():
        if file.suffix.lower() in extensions:
            yield file


def __join_label_and_image_files(
    images: Iterable[Path], labels: Iterable[Path]
) -> Iterator[tuple[Path, Path]]:
    lookup = {path.stem: path for path in images}
    for label_path in labels:
        img_path = lookup.get(label_path.stem)
        if img_path is None:
            continue
        yield (img_path, label_path)


def from_dir(
    img_dir: str | Path,
    label_dir: str | Path,
    deserializer: LabelImageDeserializer,
) -> Iterator[LabelImage]:
    """
    Generate an iterator of LabelImage objects from a directory containing images and labels.

    Args:
        img_dir: The directory containing the images.
        label_dir: The directory containing the labels.
        deserializer: A LabelImageDeserializer instance used to deserialize the label files.

    Returns:
        Iterator[LabelImage]: Iterator over LabelImage instances.
    """
    if isinstance(img_dir, str):
        img_dir = Path(img_dir)
    if isinstance(label_dir, str):
        label_dir = Path(label_dir)

    if not img_dir.exists() or not label_dir.exists():
        return DeserializedLabelImageIterator(iter([]), deserializer)

    image_paths = __get_files_with_extensions(img_dir, {".jpg", ".jpeg", ".png"})
    label_paths = __get_files_with_extensions(label_dir, {deserializer.file_extension})
    pairs = __join_label_and_image_files(image_paths, label_paths)

    return DeserializedLabelImageIterator(pairs, deserializer)


def from_pascal_dir(img_dir: str | Path, label_dir: str | Path) -> Iterator[LabelImage]:
    """
    Generate an iterator of LabelImage objects from a directory containing images and labels in Pascal VOC format.

    Args:
        img_dir: The directory containing the images.
        label_dir: The directory containing the labels.

    Returns:
        Iterator[LabelImage]: Iterator over LabelImage instances.
    """
    return from_dir(img_dir, label_dir, Pascal(label_dir))


def from_darknet_dir(
    img_dir: str | Path,
    label_dir: str | Path,
    class_file: str | Path | None = None,
    strict: bool = False,
) -> Iterator[LabelImage]:
    """
    Load Darknet format labels and images from directories.

    Args:
        img_dir (str | Path): Directory containing image files.
        label_dir (str | Path): Directory containing Darknet format label files.
        class_file (str | Path | None, optional): Path to the class file. If not provided,
            attempts to find 'classes.txt', 'names.txt', or '.names' in the label directory.
        strict (bool, optional): If True, raise an error if no class file is found.
            Otherwise returns an empty iterator.

    Returns:
        Iterator[LabelImage]: Iterator over LabelImage instances.
    """
    if isinstance(label_dir, str):
        label_dir = Path(label_dir)
    if isinstance(class_file, str):
        class_file = Path(class_file)
    if class_file is None:
        candidates = ["classes.txt", "names.txt", ".names"]

        for candidate in candidates:
            class_file = label_dir.joinpath(candidate)
            if class_file.exists():
                break

    if class_file is None or not class_file.exists():
        if strict:
            raise ClassFileNotFoundError(label_dir.as_posix() + " | ".join(candidates))
        logger.debug("No Darknet class file found.")
        return iter([])

    classes = Darknet.load_class_file(class_file)
    return from_dir(img_dir, label_dir, Darknet(classes))


def write(filepath: str | Path, label_images: Iterable[LabelImage]) -> None:
    """
    Writes a list of LabelImage objects to a file in AAU JSON format.

    Args:
        filepath (str | Path): The path to the file where the LabelImage objects will be written.
        label_images (Iterable[LabelImage]): The list of LabelImage objects to be written.

    Returns:
        None
    """
    label_image_dicts = map(utilities.label_image_to_dict, label_images)
    with open(filepath, "w") as file:
        json.dump(list(label_image_dicts), file, indent=2)


def __get_unique_classifiers(label_images: Iterable[LabelImage]) -> dict[str, int]:
    classifiers = set()
    for label_image in label_images:
        classifiers.update(label.name for label in label_image.labels)
    return {classifier: i for i, classifier in enumerate(classifiers)}


def to_coco(
    label_images: Iterable[LabelImage], license: COCOLicense, info: COCOInfo
) -> dict[str, Any]:
    """
    Converts a list of LabelImage objects into a COCO dataset dictionary.

    Args:
        label_images (Iterable[LabelImage]): A list of LabelImage objects.
        license (COCOLicense): The license information for the dataset.
        info (COCOInfo): The info section of the COCO dataset.

    Returns:
        dict[str, Any]: A dictionary representing the COCO dataset.
    """
    logger.info("Extracting unique labels from LabelImage objects")
    classifier_db = __get_unique_classifiers(label_images)

    categories = [
        {"id": id, "name": name, "supercategory": ""}
        for name, id in classifier_db.items()
    ]

    images = []
    annotations = []

    label_id = 0
    for image_id, label_image in enumerate(label_images):
        images.append(
            {
                "id": image_id,
                "width": label_image.width,
                "height": label_image.height,
                "file_name": Path(label_image.path).name,
                "license": 0,
                "flickr_url": "",
                "coco_url": label_image.path,
                "date_captured": datetime.now(),
            }
        )
        for label in label_image.labels:
            annotations.append(
                {
                    "id": label_id,
                    "image_id": image_id,
                    "category_id": classifier_db[label.name],
                    "bbox": [label.x, label.y, label.width, label.height],
                    "iscrowd": 0,
                }
            )
            label_id += 1

    return {
        "info": dataclasses.asdict(info),
        "images": images,
        "annotations": annotations,
        "categories": categories,
        "licenses": [
            {
                "id": 0,
                "name": license.name,
                "url": license.url,
            }
        ],
    }


def to_dataframe(label_images: Sequence[LabelImage]) -> DataFrame:
    """
    Converts a sequence of LabelImage objects into a pandas DataFrame.

    Parameters:
    label_images (Sequence[LabelImage]): A sequence of LabelImage objects.

    Returns:
    DataFrame: A pandas DataFrame containing the label image data.
    """
    rows = []
    for label_image in label_images:
        rows.extend(
            {
                "path": label_image.path,
                "image_width": label_image.width,
                "image_height": label_image.height,
                "label_width": label.width,
                "label_height": label.height,
                "x": label.x,
                "y": label.y,
                "classifier": label.name,
            }
            for label in label_image.labels
        )
    return DataFrame(rows)


def from_dataframe(df: DataFrame) -> Iterable[LabelImage]:
    """
    Converts a pandas DataFrame with the columns ["path", "image_width", "image_height", "x", "y", "label_width", "label_height", "classifier"]
    into an iterable of LabelImage objects.

    Parameters:
    df (DataFrame): A pandas DataFrame containing image data.

    Returns:
    Iterable[LabelImage]: An iterable of LabelImage objects.
    """

    # Old format uses 'classifier', new format uses 'name'
    if "classifier" in df:
        label_name = "classifier"
    else:
        label_name = "name"

    group_by = df.groupby(["path", "image_width", "image_height"])

    for (path, width, height), group in group_by:  # type: ignore
        labels: Sequence[Label] = [
            AAULabel(
                label["x"],  # type: ignore
                label["y"],  # type: ignore
                label["label_width"],  # type: ignore
                label["label_height"],  # type: ignore
                label[label_name],  # type: ignore
            )
            for _, label in group.iterrows()
        ]
        yield AAULabelImage(
            Path(path),
            width,
            height,
            labels,
        )  # type: ignore



from typing import Any

from .protocols import LabelImage


def label_image_to_dict(label_image: LabelImage) -> dict[str, Any]:
    return {
        "path": label_image.path.as_posix(),
        "width": label_image.width,
        "height": label_image.height,
        "labels": [
            {
                "x": label.x,
                "y": label.y,
                "width": label.width,
                "height": label.height,
                "name": label.name,
            }
            for label in label_image.labels
        ],
    }



from pathlib import Path

import pytest
from pytest import TempdirFactory

import aau_label

DATA_DIR = Path(__file__).parent.joinpath("data")


@pytest.fixture
def gomap_format_path():
    for path in DATA_DIR.joinpath("gomap_format").glob("*.json"):
        yield path


def test_from_file(gomap_format_path: Path):
    label_images = aau_label.from_file(gomap_format_path)
    assert isinstance(label_images, list), "label_images is not a list"


def test_write(gomap_format_path: Path, tmpdir_factory: TempdirFactory):
    file_name = Path(gomap_format_path.name)
    out = tmpdir_factory.mktemp("gomap_format_out").join(file_name)

    label_images = aau_label.from_file(gomap_format_path)
    aau_label.write(out.strpath, label_images)
    out.remove()



import os
import xml.etree.ElementTree as ET
from pathlib import Path
from typing import Union
from xml.dom import minidom
from xml.etree import ElementTree
from xml.etree.ElementTree import Element, ParseError

import PIL
import PIL.Image
from PIL.Image import Image

from ..errors import PascalParseError
from ..model import AAULabel, AAULabelImage, BoundingBox
from ..protocols import Label, LabelImage, LabelImageDeserializer, LabelImageSerializer


class Pascal(LabelImageDeserializer, LabelImageSerializer):
    file_extension = ".xml"

    def __init__(self, label_dir: Union[str, Path], indent: int = 4) -> None:
        self.label_dir = Path(label_dir) if isinstance(label_dir, str) else label_dir
        self.indent = indent

    def serialize(self, label_img: LabelImage) -> str:
        annotation = ET.Element("annotation")

        rel_img_path = os.path.relpath(label_img.path, self.label_dir)
        ET.SubElement(annotation, "folder").text = label_img.path.parent.name
        ET.SubElement(annotation, "filename").text = label_img.path.name
        ET.SubElement(annotation, "path").text = rel_img_path

        self.__add_source_element(annotation, label_img)
        self.__add_size_element(annotation, label_img)
        self.__add_segmented_element(annotation)
        self.__add_object_elements(annotation, label_img)

        return self.__prettify(annotation)

    def __deserialize(self, image_file: Path, label_file: Path, image: Image):
        try:
            root = ElementTree.parse(label_file).getroot()
            width, height = image.size
            labels = [
                self.__parse_object(child) for child in root if child.tag == "object"
            ]
            return AAULabelImage(image_file, width, height, labels)
        except ParseError as e:
            raise PascalParseError(label_file) from e

    def deserialize(
        self, image_file: Path, label_file: Path, image: Image | None = None
    ) -> LabelImage:
        if image is None:
            image = PIL.Image.open(image_file)
            return self.__deserialize(image_file, label_file, image)
        return self.__deserialize(image_file, label_file, image)

    def __prettify(self, elem: Element):
        """Return a pretty-printed XML string for the Element."""
        rough_string = ET.tostring(elem, "unicode")
        reparsed = minidom.parseString(rough_string)
        return reparsed.toprettyxml(indent=" " * self.indent)

    def __add_source_element(self, element: Element, label_img: LabelImage) -> None:
        source = ET.SubElement(element, "source")
        credit = "Unknown"
        ET.SubElement(source, "database").text = credit

    def __add_size_element(self, element: Element, label_img: LabelImage) -> None:
        size = ET.SubElement(element, "size")
        ET.SubElement(size, "width").text = str(label_img.width)
        ET.SubElement(size, "height").text = str(label_img.height)
        ET.SubElement(size, "depth").text = str(3)

    def __add_segmented_element(self, element: Element) -> None:
        ET.SubElement(element, "segmented").text = str(0)

    def __add_object_elements(self, element: Element, label_img: LabelImage) -> None:
        for label in label_img.labels:
            object_element = ET.SubElement(element, "object")
            ET.SubElement(object_element, "name").text = label.name
            ET.SubElement(object_element, "pose").text = "Unspecified"
            ET.SubElement(object_element, "truncated").text = str(0)
            ET.SubElement(object_element, "difficult").text = str(0)

            bndbox = ET.SubElement(object_element, "bndbox")
            ET.SubElement(bndbox, "xmin").text = str(label.x)
            ET.SubElement(bndbox, "ymin").text = str(label.y)
            ET.SubElement(bndbox, "xmax").text = str(label.x + label.width)
            ET.SubElement(bndbox, "ymax").text = str(label.y + label.height)

    def __parse_object(self, element: Element) -> Label:
        classifier: str
        bounding_box: BoundingBox

        for child in element:
            if child.tag == "name":
                classifier = child.text or ""
            elif child.tag == "bndbox":
                bounding_box = self.__parse_bounding_box(child)

        return AAULabel(
            bounding_box.xmin,
            bounding_box.ymin,
            bounding_box.xmax - bounding_box.xmin,
            bounding_box.ymax - bounding_box.ymin,
            classifier,
        )

    def __parse_bounding_box(self, element: Element):
        xmin: int
        ymin: int
        xmax: int
        ymax: int

        for child in element:
            if child.tag not in {"xmin", "ymin", "xmax", "ymax"}:
                continue
            if not child.text:
                raise ValueError(f"{child.tag} is missing")

            if child.tag == "xmin":
                xmin = int(child.text)
            elif child.tag == "ymin":
                ymin = int(child.text)
            elif child.tag == "xmax":
                xmax = int(child.text)
            elif child.tag == "ymax":
                ymax = int(child.text)

        return BoundingBox(xmin, ymin, xmax, ymax)



import logging
from pathlib import Path
from typing import Sequence

import PIL
import PIL.Image
from PIL.Image import Image

from ..errors import ClassFileNotFoundError, DarknetParseError
from ..model import AAULabel, AAULabelImage
from ..protocols import Label, LabelImage, LabelImageDeserializer, LabelImageSerializer


class Darknet(LabelImageDeserializer, LabelImageSerializer):
    file_extension = ".txt"

    def __init__(self, labels: Sequence[str]) -> None:
        self.__labels = labels
        self.__class_map = {label: i for i, label in enumerate(labels)}
        self.__logger = logging.getLogger(__class__.__name__)

    @staticmethod
    def load_class_file(class_file: Path) -> Sequence[str]:
        try:
            with open(class_file, "r") as file:
                return [row.strip() for row in file if row]
        except FileNotFoundError as e:
            raise ClassFileNotFoundError(class_file) from e

    def serialize(self, label_img: LabelImage) -> str:
        rv = ""

        for label in label_img.labels:
            class_id = self.__class_map.get(label.name)
            if class_id is None:
                self.__logger.warning(f"{label.name} not in provided class map")
                continue

            x = (label.x + label.width / 2) / label_img.width
            y = (label.y + label.height / 2) / label_img.height
            width = label.width / label_img.width
            height = label.height / label_img.height
            rv += f"{class_id} {x} {y} {width} {height}\n"

        return rv

    def __deserialize(self, image_file: Path, label_file: Path, image: Image):
        with open(label_file, "r") as file:
            width, height = image.size
            try:
                labels = [self.__parse_row(row, width, height) for row in file]
                return AAULabelImage(image_file, width, height, labels)
            except ValueError:
                raise DarknetParseError(label_file)

    def deserialize(
        self, image_file: Path, label_file: Path, image: Image | None = None
    ) -> LabelImage:
        if image is None:
            with PIL.Image.open(image_file) as image:
                return self.__deserialize(image_file, label_file, image)
        return self.__deserialize(image_file, label_file, image)

    def __parse_row(self, row: str, img_width, img_height) -> Label:
        class_index, x, y, width, height = row.split()

        class_index = int(class_index)
        width = round(float(width) * img_width)
        height = round(float(height) * img_height)
        x = round(float(x) * img_width - width / 2)
        y = round(float(y) * img_height - height / 2)

        return AAULabel(x, y, width, height, self.__labels[class_index])



from dataclasses import dataclass
from pathlib import Path
from typing import Sequence

from .protocols import Label


@dataclass
class AAULabel:
    x: int
    y: int
    width: int
    height: int
    name: str


@dataclass
class AAULabelImage:
    path: Path
    width: int
    height: int
    labels: Sequence[Label]
    source: str | None = None


@dataclass
class COCOLicense:
    name: str
    url: str


@dataclass
class COCOInfo:
    year: int
    version: str
    description: str
    contributor: str
    url: str
    date_created: str


@dataclass
class BoundingBox:
    xmin: int
    ymin: int
    xmax: int
    ymax: int



from pathlib import Path


class DeserializerError(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class ClassFileNotFoundError(DeserializerError):
    def __init__(self, path: Path | str) -> None:
        super().__init__(f"Could not find {path}")


class PascalParseError(DeserializerError):
    def __init__(self, path: Path | str) -> None:
        super().__init__(f"Could not parse {path} as Pascal format")


class DarknetParseError(DeserializerError):
    def __init__(self, path: Path | str) -> None:
        super().__init__(f"Could not parse {path} as Darknet format")



from pathlib import Path
from typing import ClassVar, Protocol, Sequence

from PIL.Image import Image
from typing_extensions import deprecated


class Label(Protocol):
    x: int
    y: int
    width: int
    height: int
    name: str


class LabelImage(Protocol):
    path: Path
    width: int
    height: int
    labels: Sequence[Label]


@deprecated("Use LabelImageDeserializer instead")
class LabelParser(Protocol):
    file_extension: ClassVar[str]

    def parse(self, label_file: Path, image: Image) -> Sequence[Label]:
        raise NotImplementedError


class LabelImageDeserializer(Protocol):
    file_extension: ClassVar[str]

    def deserialize(
        self, img_file: Path, label_file: Path, image: Image | None = None
    ) -> LabelImage:
        raise NotImplementedError


class LabelImageSerializer(Protocol):
    def serialize(self, label: LabelImage) -> str:
        raise NotImplementedError



from __future__ import annotations

import os
from pathlib import Path
from typing import List, Optional

from PIL.Image import Image
from typing_extensions import deprecated

from ..errors import ClassFileNotFoundError
from ..model import AAULabel
from ..protocols import Label, LabelParser


@deprecated("Use aau_label.io.Darknet instead")
class DarknetParser(LabelParser):
    file_extension = ".txt"

    def __init__(self):
        self.current_directory: Optional[str] = None
        self.label_map: List[str] = []

    def __parse_row(self, row: str, img_width, img_height) -> Label:
        class_index, x, y, width, height = row.split()

        class_index = int(class_index)
        width = round(float(width) * img_width)
        height = round(float(height) * img_height)
        x = round(float(x) * img_width - width / 2)
        y = round(float(y) * img_height - height / 2)

        return AAULabel(x, y, width, height, self.label_map[class_index])

    def parse(self, label_file: Path, image: Image) -> List[Label]:
        self.__load_class_file(label_file)

        with open(label_file, "r") as file:
            width, height = image.size

            try:
                return [self.__parse_row(row, width, height) for row in file]
            except ValueError:
                raise ValueError(f"Badly formatted label file: {label_file}")

    def __load_class_file(self, label_file: Path):
        filepath, _ = os.path.split(label_file)

        if self.current_directory == filepath:
            return

        self.current_directory = filepath
        classfile = os.path.join(filepath, "classes.txt")

        try:
            with open(classfile, "r") as file:
                self.label_map = [row.strip() for row in file if row]
        except FileNotFoundError as e:
            raise ClassFileNotFoundError(filepath) from e



from dataclasses import dataclass
from pathlib import Path
from typing import List
from xml.etree import ElementTree
from xml.etree.ElementTree import Element, ParseError

from PIL.Image import Image
from typing_extensions import deprecated

from aau_label.errors import PascalParseError

from ..model import AAULabel
from ..protocols import Label, LabelParser


@dataclass
class BoundingBox:
    xmin: int
    ymin: int
    xmax: int
    ymax: int


@deprecated("Use aau_label.io.Pascal instead")
class PascalParser(LabelParser):
    file_extension = ".xml"

    def parse(self, label_file: Path, image: Image) -> List[Label]:
        root = ElementTree.parse(label_file).getroot()
        try:
            return [
                self.__parse_object(child) for child in root if child.tag == "object"
            ]
        except ParseError as e:
            raise PascalParseError(label_file) from e

    def __parse_object(self, element: Element) -> Label:
        classifier: str
        bounding_box: BoundingBox

        for child in element:
            if child.tag == "name":
                classifier = child.text or ""
            elif child.tag == "bndbox":
                bounding_box = self.__parse_bounding_box(child)

        return AAULabel(
            bounding_box.xmin,
            bounding_box.ymin,
            bounding_box.xmax - bounding_box.xmin,
            bounding_box.ymax - bounding_box.ymin,
            classifier,
        )

    def __parse_bounding_box(self, element: Element):
        xmin: int
        ymin: int
        xmax: int
        ymax: int

        for child in element:
            if child.tag not in {"xmin", "ymin", "xmax", "ymax"}:
                continue
            if not child.text:
                raise ValueError(f"{child.tag} is missing")

            if child.tag == "xmin":
                xmin = int(child.text)
            elif child.tag == "ymin":
                ymin = int(child.text)
            elif child.tag == "xmax":
                xmax = int(child.text)
            elif child.tag == "ymax":
                ymax = int(child.text)

        return BoundingBox(xmin, ymin, xmax, ymax)
